@@ -2,7 +2,6 @@
  * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2007,
  *               2010, 2011 Free Software Foundation, Inc.
  * Copyright (C) 2014, 2019 Werner Koch
- * Copyright (C) 2023 g10 Code GmbH
  *
  * This file is part of GnuPG.
  *
@@ -40,13 +39,6 @@
 #define O_BINARY 0
 #endif
 
-
-static gpg_error_t read_key_file (const unsigned char *grip,
-                                  gcry_sexp_t *result, nvc_t *r_keymeta,
-                                  char **r_orig_key_value);
-static gpg_error_t is_shadowed_key (gcry_sexp_t s_skey);
-
-
 /* Helper to pass data to the check callback of the unprotect function. */
 struct try_unprotect_arg_s
 {
@@ -58,158 +50,98 @@ struct try_unprotect_arg_s
 };
 
 
-/* Return the file name for the 20 byte keygrip GRIP.  With FOR_NEW
- * create a file name for later renaming to the actual name.  Return
- * NULL on error. */
+/* Repalce all linefeeds in STRING by "%0A" and return a new malloced
+ * string.  May return NULL on memory error.  */
 static char *
-fname_from_keygrip (const unsigned char *grip, int for_new)
+linefeed_to_percent0A (const char *string)
 {
-  char hexgrip[40+4+4+1];
+  const char *s;
+  size_t n;
+  char *buf, *p;
 
-  bin2hex (grip, 20, hexgrip);
-  strcpy (hexgrip+40, for_new? ".key.tmp" : ".key");
-
-  return make_filename_try (gnupg_homedir (), GNUPG_PRIVATE_KEYS_DIR,
-                            hexgrip, NULL);
+  for (n=0, s=string; *s; s++)
+    if (*s == '\n')
+      n += 3;
+    else
+      n++;
+  p = buf = xtrymalloc (n+1);
+  if (!buf)
+    return NULL;
+  for (s=string; *s; s++)
+    if (*s == '\n')
+      {
+        memcpy (p, "%0A", 3);
+        p += 3;
+      }
+    else
+      *p++ = *s;
+  *p = 0;
+  return buf;
 }
 
 
-/* Write the S-expression formatted key (BUFFER,LENGTH) to our key
- * storage.  With FORCE passed as true an existing key with the given
- * GRIP will be overwritten.  If SERIALNO and KEYREF are given a Token
- * line is added to the key if the extended format is used.  If
- * TIMESTAMP is not zero and the key doies not yet exists it will be
- * recorded as creation date.  */
-int
-agent_write_private_key (const unsigned char *grip,
-                         const void *buffer, size_t length, int force,
-                         const char *serialno, const char *keyref,
-                         const char *dispserialno,
-                         time_t timestamp)
+/* Note: Ownership of FNAME and FP are moved to this function.  */
+static gpg_error_t
+write_extended_private_key (char *fname, estream_t fp, int update, int newkey,
+                            const void *buf, size_t len,
+                            const char *serialno, const char *keyref,
+                            time_t timestamp)
 {
   gpg_error_t err;
-  char *oldfname = NULL;
-  char *fname = NULL;
-  estream_t fp;
-  int newkey = 0;
   nvc_t pk = NULL;
   gcry_sexp_t key = NULL;
-  int is_regular;
   int remove = 0;
-  char *token0 = NULL;
   char *token = NULL;
-  char *dispserialno_buffer = NULL;
-  char **tokenfields = NULL;
-  int blocksigs = 0;
-  char *orig_key_value = NULL;
-  const char *s;
-  int force_modify = 0;
 
-  oldfname = fname_from_keygrip (grip, 0);
-  if (!oldfname)
-    return out_of_core ();
-
-  err = read_key_file (grip, &key, &pk, &orig_key_value);
-  if (err)
+  if (update)
     {
-      if (gpg_err_code (err) == GPG_ERR_ENOENT)
-        newkey = 1;
-      else
+      int line;
+
+      err = nvc_parse_private_key (&pk, &line, fp);
+      if (err && gpg_err_code (err) != GPG_ERR_ENOENT)
         {
-          log_error ("can't open '%s': %s\n", oldfname, gpg_strerror (err));
+          log_error ("error parsing '%s' line %d: %s\n",
+                     fname, line, gpg_strerror (err));
           goto leave;
         }
     }
-
-  nvc_modified (pk, 1);  /* Clear that flag after a read.  */
-
-  if (!pk)
+  else
     {
-      /* Key is still in the old format or does not exist - create a
-       * new container.  */
       pk = nvc_new_private_key ();
       if (!pk)
         {
           err = gpg_error_from_syserror ();
           goto leave;
         }
-      force_modify = 1;
     }
+  es_clearerr (fp);
 
-  /* Check whether we already have a regular key.  */
-  is_regular = (key && gpg_err_code (is_shadowed_key (key)) != GPG_ERR_TRUE);
-
-  /* Turn (BUFFER,LENGTH) into a gcrypt s-expression and set it into
-   * our name value container.  */
-  gcry_sexp_release (key);
-  err = gcry_sexp_sscan (&key, NULL, buffer, length);
+  err = gcry_sexp_sscan (&key, NULL, buf, len);
   if (err)
     goto leave;
+
   err = nvc_set_private_key (pk, key);
   if (err)
     goto leave;
 
-  /* Detect whether the key value actually changed and if not clear
-   * the modified flag.  This extra check is required because
-   * read_key_file removes the Key entry from the container and we
-   * then create a new Key entry which might be the same, though.  */
-  if (!force_modify
-      && orig_key_value && (s = nvc_get_string (pk, "Key:"))
-      && !strcmp (orig_key_value, s))
-    {
-      nvc_modified (pk, 1);  /* Clear that flag.  */
-    }
-  xfree (orig_key_value);
-  orig_key_value = NULL;
-
-  /* Check that we do not update a regular key with a shadow key.  */
-  if (is_regular && gpg_err_code (is_shadowed_key (key)) == GPG_ERR_TRUE)
-    {
-      log_info ("updating regular key file '%s'"
-                " by a shadow key inhibited\n", oldfname);
-      err = 0;  /* Simply ignore the error.  */
-      goto leave;
-    }
-  /* Check that we update a regular key only in force mode.  */
-  if (is_regular && !force)
-    {
-      log_error ("secret key file '%s' already exists\n", oldfname);
-      err = gpg_error (GPG_ERR_EEXIST);
-      goto leave;
-    }
-
   /* If requested write a Token line.  */
   if (serialno && keyref)
     {
       nve_t item;
-      size_t token0len;
+      const char *s;
 
-      if (dispserialno)
-        {
-          /* Escape the DISPSERIALNO.  */
-          dispserialno_buffer = percent_plus_escape (dispserialno);
-          if (!dispserialno_buffer)
-            {
-              err = gpg_error_from_syserror ();
-              goto leave;
-            }
-          dispserialno = dispserialno_buffer;
-        }
-
-      token0 = strconcat (serialno, " ", keyref, NULL);
-      if (token0)
-        token = strconcat (token0, " - ", dispserialno? dispserialno:"-", NULL);
-      if (!token0 || !token)
+      token = strconcat (serialno, " ", keyref, NULL);
+      if (!token)
         {
           err = gpg_error_from_syserror ();
           goto leave;
         }
 
-      token0len = strlen (token0);
+      /* fixme: the strcmp should compare only the first two strings.  */
       for (item = nvc_lookup (pk, "Token:");
            item;
            item = nve_next_value (item, "Token:"))
-        if ((s = nve_value (item)) && !strncmp (s, token0, token0len))
+        if ((s = nve_value (item)) && !strcmp (s, token))
           break;
       if (!item)
         {
@@ -220,25 +152,9 @@ agent_write_private_key (const unsigned char *grip,
           if (err)
             goto leave;
         }
-      else
-        {
-          /* Token exists: Update the display s/n.  It may have
-           * changed due to changes in a newer software version.  */
-          if (s && (tokenfields = strtokenize (s, " \t\n"))
-              && tokenfields[0] && tokenfields[1] && tokenfields[2]
-              && tokenfields[3]
-              && !strcmp (tokenfields[3], dispserialno))
-            ; /* No need to update Token entry.  */
-          else
-            {
-              err = nve_set (pk, item, token);
-              if (err)
-                goto leave;
-            }
-        }
     }
 
-  /* If a timestamp has been supplied and the key is new, write a
+  /* If a timestamp has been supplied and the key is new write a
    * creation timestamp.  (We douple check that there is no Created
    * item yet.)*/
   if (timestamp && newkey && !nvc_lookup (pk, "Created:"))
@@ -251,26 +167,12 @@ agent_write_private_key (const unsigned char *grip,
         goto leave;
     }
 
-  /* Check whether we need to write the file at all.  */
-  if (!nvc_modified (pk, 0))
-    {
-      err = 0;
-      goto leave;
-    }
 
-  /* Create a temporary file for writing.  */
-  fname = fname_from_keygrip (grip, 1);
-  fp = fname ? es_fopen (fname, "wbx,mode=-rw") : NULL;
-  if (!fp)
-    {
-      err = gpg_error_from_syserror ();
-      log_error ("can't create '%s': %s\n", fname, gpg_strerror (err));
-      goto leave;
-    }
+  err = es_fseek (fp, 0, SEEK_SET);
+  if (err)
+    goto leave;
 
   err = nvc_write (pk, fp);
-  if (!err && es_fflush (fp))
-    err = gpg_error_from_syserror ();
   if (err)
     {
       log_error ("error writing '%s': %s\n", fname, gpg_strerror (err));
@@ -278,6 +180,14 @@ agent_write_private_key (const unsigned char *grip,
       goto leave;
     }
 
+  if (ftruncate (es_fileno (fp), es_ftello (fp)))
+    {
+      err = gpg_error_from_syserror ();
+      log_error ("error truncating '%s': %s\n", fname, gpg_strerror (err));
+      remove = 1;
+      goto leave;
+    }
+
   if (es_fclose (fp))
     {
       err = gpg_error_from_syserror ();
@@ -285,37 +195,149 @@ agent_write_private_key (const unsigned char *grip,
       remove = 1;
       goto leave;
     }
-  fp = NULL;
-
-  err = gnupg_rename_file (fname, oldfname, &blocksigs);
-  if (err)
-    {
-      err = gpg_error_from_syserror ();
-      log_error ("error renaming '%s': %s\n", fname, gpg_strerror (err));
-      remove = 1;
-      goto leave;
-    }
+  else
+    fp = NULL;
 
   bump_key_eventcounter ();
 
  leave:
-  if (blocksigs)
-    gnupg_unblock_all_signals ();
   es_fclose (fp);
-  if (remove && fname)
+  if (remove)
     gnupg_remove (fname);
-  xfree (orig_key_value);
   xfree (fname);
-  xfree (oldfname);
-  xfree (token);
-  xfree (token0);
-  xfree (dispserialno_buffer);
-  xfree (tokenfields);
   gcry_sexp_release (key);
   nvc_release (pk);
+  xfree (token);
   return err;
 }
 
+/* Write an S-expression formatted key to our key storage.  With FORCE
+ * passed as true an existing key with the given GRIP will get
+ * overwritten.  If SERIALNO and KEYREF are given a Token line is
+ * added to the key if the extended format is used.  If TIMESTAMP is
+ * not zero and the key doies not yet exists it will be recorded as
+ * creation date.  */
+int
+agent_write_private_key (const unsigned char *grip,
+                         const void *buffer, size_t length, int force,
+                         const char *serialno, const char *keyref,
+                         time_t timestamp)
+{
+  char *fname;
+  estream_t fp;
+  char hexgrip[40+4+1];
+
+  bin2hex (grip, 20, hexgrip);
+  strcpy (hexgrip+40, ".key");
+
+  fname = make_filename (gnupg_homedir (), GNUPG_PRIVATE_KEYS_DIR,
+                         hexgrip, NULL);
+
+  /* FIXME: Write to a temp file first so that write failures during
+     key updates won't lead to a key loss.  */
+
+  if (!force && !gnupg_access (fname, F_OK))
+    {
+      log_error ("secret key file '%s' already exists\n", fname);
+      xfree (fname);
+      return gpg_error (GPG_ERR_EEXIST);
+    }
+
+  fp = es_fopen (fname, force? "rb+,mode=-rw" : "wbx,mode=-rw");
+  if (!fp)
+    {
+      gpg_error_t tmperr = gpg_error_from_syserror ();
+
+      if (force && gpg_err_code (tmperr) == GPG_ERR_ENOENT)
+        {
+          fp = es_fopen (fname, "wbx,mode=-rw");
+          if (!fp)
+            tmperr = gpg_error_from_syserror ();
+        }
+      if (!fp)
+        {
+          log_error ("can't create '%s': %s\n", fname, gpg_strerror (tmperr));
+          xfree (fname);
+          return tmperr;
+        }
+    }
+  else if (force)
+    {
+      gpg_error_t rc;
+      char first;
+
+      /* See if an existing key is in extended format.  */
+      if (es_fread (&first, 1, 1, fp) != 1)
+        {
+          rc = gpg_error_from_syserror ();
+          log_error ("error reading first byte from '%s': %s\n",
+                     fname, strerror (errno));
+          xfree (fname);
+          es_fclose (fp);
+          return rc;
+        }
+
+      rc = es_fseek (fp, 0, SEEK_SET);
+      if (rc)
+        {
+          log_error ("error seeking in '%s': %s\n", fname, strerror (errno));
+          xfree (fname);
+          es_fclose (fp);
+          return rc;
+        }
+
+      if (first != '(')
+        {
+          /* Key is already in the extended format.  */
+          return write_extended_private_key (fname, fp, 1, 0, buffer, length,
+                                             serialno, keyref, timestamp);
+        }
+      if (first == '(' && opt.enable_extended_key_format)
+        {
+          /* Key is in the old format - but we want the extended format.  */
+          return write_extended_private_key (fname, fp, 0, 0, buffer, length,
+                                             serialno, keyref, timestamp);
+        }
+    }
+
+  if (opt.enable_extended_key_format)
+    return write_extended_private_key (fname, fp, 0, 1, buffer, length,
+                                       serialno, keyref, timestamp);
+
+  if (es_fwrite (buffer, length, 1, fp) != 1)
+    {
+      gpg_error_t tmperr = gpg_error_from_syserror ();
+      log_error ("error writing '%s': %s\n", fname, gpg_strerror (tmperr));
+      es_fclose (fp);
+      gnupg_remove (fname);
+      xfree (fname);
+      return tmperr;
+    }
+
+  /* When force is given, the file might have to be truncated.  */
+  if (force && ftruncate (es_fileno (fp), es_ftello (fp)))
+    {
+      gpg_error_t tmperr = gpg_error_from_syserror ();
+      log_error ("error truncating '%s': %s\n", fname, gpg_strerror (tmperr));
+      es_fclose (fp);
+      gnupg_remove (fname);
+      xfree (fname);
+      return tmperr;
+    }
+
+  if (es_fclose (fp))
+    {
+      gpg_error_t tmperr = gpg_error_from_syserror ();
+      log_error ("error closing '%s': %s\n", fname, gpg_strerror (tmperr));
+      gnupg_remove (fname);
+      xfree (fname);
+      return tmperr;
+    }
+  bump_key_eventcounter ();
+  xfree (fname);
+  return 0;
+}
+
 
 /* Callback function to try the unprotection from the passphrase query
    code. */
@@ -394,6 +416,33 @@ try_unprotect_cb (struct pin_entry_info_s *pi)
 }
 
 
+/* Return true if the STRING has an %C or %c expando.  */
+static int
+has_comment_expando (const char *string)
+{
+  const char *s;
+  int percent = 0;
+
+  if (!string)
+    return 0;
+
+  for (s = string; *s; s++)
+    {
+      if (percent)
+        {
+          if (*s == 'c' || *s == 'C')
+            return 1;
+          percent = 0;
+        }
+      else if (*s == '%')
+        percent = 1;
+    }
+  return 0;
+}
+
+
+
+
 /* Modify a Key description, replacing certain special format
    characters.  List of currently supported replacements:
 
@@ -672,7 +721,7 @@ unprotect (ctrl_t ctrl, const char *cache_nonce, const char *desc_text,
             {
               /* We need to give the other thread a chance to actually put
                  it into the cache. */
-              npth_sleep (1);
+              gnupg_sleep (1);
               goto retry;
             }
           /* Timeout - better call pinentry now the plain way. */
@@ -758,43 +807,40 @@ unprotect (ctrl_t ctrl, const char *cache_nonce, const char *desc_text,
 
 /* Read the key identified by GRIP from the private key directory and
  * return it as an gcrypt S-expression object in RESULT.  If R_KEYMETA
- * is not NULL, the meta data items are stored there.  However the
- * "Key:" item is removed.  If R_ORIG_KEY_VALUE is non-NULL and the
- * Key items was removed, its value is stored at that R_ORIG_KEY_VALUE
- * and the caller must free it.  Returns an error code and stores NULL
- * at RESULT. */
+ * is not NULl and the extended key format is used, the meta data
+ * items are stored there.  However the "Key:" item is removed from
+ * it.  On failure returns an error code and stores NULL at RESULT and
+ * R_KEYMETA. */
 static gpg_error_t
-read_key_file (const unsigned char *grip, gcry_sexp_t *result, nvc_t *r_keymeta,
-               char **r_orig_key_value)
+read_key_file (const unsigned char *grip, gcry_sexp_t *result, nvc_t *r_keymeta)
 {
   gpg_error_t err;
   char *fname;
-  estream_t fp = NULL;
+  estream_t fp;
   struct stat st;
-  unsigned char *buf = NULL;
+  unsigned char *buf;
   size_t buflen, erroff;
   gcry_sexp_t s_skey;
+  char hexgrip[40+4+1];
   char first;
 
   *result = NULL;
   if (r_keymeta)
     *r_keymeta = NULL;
-  if (r_orig_key_value)
-    *r_orig_key_value = NULL;
 
-  fname = fname_from_keygrip (grip, 0);
-  if (!fname)
-    {
-      err = gpg_error_from_syserror ();
-      goto leave;
-    }
+  bin2hex (grip, 20, hexgrip);
+  strcpy (hexgrip+40, ".key");
+
+  fname = make_filename (gnupg_homedir (), GNUPG_PRIVATE_KEYS_DIR,
+                         hexgrip, NULL);
   fp = es_fopen (fname, "rb");
   if (!fp)
     {
       err = gpg_error_from_syserror ();
       if (gpg_err_code (err) != GPG_ERR_ENOENT)
         log_error ("can't open '%s': %s\n", fname, gpg_strerror (err));
-      goto leave;
+      xfree (fname);
+      return err;
     }
 
   if (es_fread (&first, 1, 1, fp) != 1)
@@ -802,23 +848,29 @@ read_key_file (const unsigned char *grip, gcry_sexp_t *result, nvc_t *r_keymeta,
       err = gpg_error_from_syserror ();
       log_error ("error reading first byte from '%s': %s\n",
                  fname, gpg_strerror (err));
-      goto leave;
+      xfree (fname);
+      es_fclose (fp);
+      return err;
     }
 
   if (es_fseek (fp, 0, SEEK_SET))
     {
       err = gpg_error_from_syserror ();
       log_error ("error seeking in '%s': %s\n", fname, gpg_strerror (err));
-      goto leave;
+      xfree (fname);
+      es_fclose (fp);
+      return err;
     }
 
   if (first != '(')
     {
       /* Key is in extended format.  */
-      nvc_t pk;
+      nvc_t pk = NULL;
       int line;
 
       err = nvc_parse_private_key (&pk, &line, fp);
+      es_fclose (fp);
+
       if (err)
         log_error ("error parsing '%s' line %d: %s\n",
                    fname, line, gpg_strerror (err));
@@ -829,37 +881,24 @@ read_key_file (const unsigned char *grip, gcry_sexp_t *result, nvc_t *r_keymeta,
             log_error ("error getting private key from '%s': %s\n",
                        fname, gpg_strerror (err));
           else
-            {
-              if (r_orig_key_value)
-                {
-                  const char *s = nvc_get_string (pk, "Key:");
-                  if (s)
-                    {
-                      *r_orig_key_value = xtrystrdup (s);
-                      if (!*r_orig_key_value)
-                        {
-                          err = gpg_error_from_syserror ();
-                          nvc_release (pk);
-                          goto leave;
-                        }
-                    }
-                }
-              nvc_delete_named (pk, "Key:");
-            }
+            nvc_delete_named (pk, "Key:");
         }
 
       if (!err && r_keymeta)
         *r_keymeta = pk;
       else
         nvc_release (pk);
-      goto leave;
+      xfree (fname);
+      return err;
     }
 
   if (fstat (es_fileno (fp), &st))
     {
       err = gpg_error_from_syserror ();
       log_error ("can't stat '%s': %s\n", fname, gpg_strerror (err));
-      goto leave;
+      xfree (fname);
+      es_fclose (fp);
+      return err;
     }
 
   buflen = st.st_size;
@@ -869,7 +908,11 @@ read_key_file (const unsigned char *grip, gcry_sexp_t *result, nvc_t *r_keymeta,
       err = gpg_error_from_syserror ();
       log_error ("error allocating %zu bytes for '%s': %s\n",
                  buflen, fname, gpg_strerror (err));
-      goto leave;
+      xfree (fname);
+      es_fclose (fp);
+      xfree (buf);
+      return err;
+
     }
 
   if (es_fread (buf, buflen, 1, fp) != 1)
@@ -877,24 +920,25 @@ read_key_file (const unsigned char *grip, gcry_sexp_t *result, nvc_t *r_keymeta,
       err = gpg_error_from_syserror ();
       log_error ("error reading %zu bytes from '%s': %s\n",
                  buflen, fname, gpg_strerror (err));
-      goto leave;
+      xfree (fname);
+      es_fclose (fp);
+      xfree (buf);
+      return err;
     }
 
   /* Convert the file into a gcrypt S-expression object.  */
   err = gcry_sexp_sscan (&s_skey, &erroff, (char*)buf, buflen);
+  xfree (fname);
+  es_fclose (fp);
+  xfree (buf);
   if (err)
     {
       log_error ("failed to build S-Exp (off=%u): %s\n",
                  (unsigned int)erroff, gpg_strerror (err));
-      goto leave;
+      return err;
     }
   *result = s_skey;
-
- leave:
-  es_fclose (fp);
-  xfree (fname);
-  xfree (buf);
-  return err;
+  return 0;
 }
 
 
@@ -937,24 +981,22 @@ agent_key_from_file (ctrl_t ctrl, const char *cache_nonce,
                      const char *desc_text,
                      const unsigned char *grip, unsigned char **shadow_info,
                      cache_mode_t cache_mode, lookup_ttl_t lookup_ttl,
-                     gcry_sexp_t *result, char **r_passphrase,
-                     uint64_t *r_timestamp)
+                     gcry_sexp_t *result, char **r_passphrase)
 {
   gpg_error_t err;
   unsigned char *buf;
-  size_t len, buflen, erroff;
+  size_t len, erroff;
   gcry_sexp_t s_skey;
   nvc_t keymeta = NULL;
+  char *desc_text_buffer = NULL;  /* Used in case we extend DESC_TEXT.  */
 
   *result = NULL;
   if (shadow_info)
     *shadow_info = NULL;
   if (r_passphrase)
     *r_passphrase = NULL;
-  if (r_timestamp)
-    *r_timestamp = (uint64_t)(-1);
 
-  err = read_key_file (grip, &s_skey, &keymeta, NULL);
+  err = read_key_file (grip, &s_skey, &keymeta);
   if (err)
     {
       if (gpg_err_code (err) == GPG_ERR_ENOENT)
@@ -969,18 +1011,10 @@ agent_key_from_file (ctrl_t ctrl, const char *cache_nonce,
   if (err)
     {
       nvc_release (keymeta);
+      xfree (desc_text_buffer);
       return err;
     }
 
-  if (r_timestamp && keymeta)
-    {
-      const char *created = nvc_get_string (keymeta, "Created:");
-
-      if (created)
-        *r_timestamp = isotime2epoch_u64 (created);
-    }
-  nvc_release (keymeta);
-
   switch (agent_private_key_type (buf))
     {
     case PRIVATE_KEY_CLEAR:
@@ -1004,25 +1038,43 @@ agent_key_from_file (ctrl_t ctrl, const char *cache_nonce,
     case PRIVATE_KEY_PROTECTED:
       {
 	char *desc_text_final;
-	char *comment = NULL;
+	char *comment_buffer = NULL;
+	const char *comment = NULL;
 
         /* Note, that we will take the comment as a C string for
-           display purposes; i.e. all stuff beyond a Nul character is
-           ignored.  */
-        {
-          gcry_sexp_t comment_sexp;
+         * display purposes; i.e. all stuff beyond a Nul character is
+         * ignored.  If a "Label" entry is available in the meta data
+         * this is used instead of the s-expression comment.  */
+        if (keymeta && (comment = nvc_get_string (keymeta, "Label:")))
+          {
+            if (strchr (comment, '\n')
+                && (comment_buffer = linefeed_to_percent0A (comment)))
+              comment = comment_buffer;
+            /* In case DESC_TEXT has no escape pattern for a comment
+             * we append one.  */
+            if (desc_text && !has_comment_expando (desc_text))
+              {
+                desc_text_buffer = strconcat (desc_text, "%0A%C", NULL);
+                if (desc_text_buffer)
+                  desc_text = desc_text_buffer;
+              }
+          }
+        else
+          {
+            gcry_sexp_t comment_sexp;
 
-          comment_sexp = gcry_sexp_find_token (s_skey, "comment", 0);
-          if (comment_sexp)
-            comment = gcry_sexp_nth_string (comment_sexp, 1);
-          gcry_sexp_release (comment_sexp);
-        }
+            comment_sexp = gcry_sexp_find_token (s_skey, "comment", 0);
+            if (comment_sexp)
+              comment_buffer = gcry_sexp_nth_string (comment_sexp, 1);
+            gcry_sexp_release (comment_sexp);
+            comment = comment_buffer;
+          }
 
         desc_text_final = NULL;
 	if (desc_text)
           err = agent_modify_description (desc_text, comment, s_skey,
                                           &desc_text_final);
-        gcry_free (comment);
+        gcry_free (comment_buffer);
 
 	if (!err)
 	  {
@@ -1077,13 +1129,15 @@ agent_key_from_file (ctrl_t ctrl, const char *cache_nonce,
           xfree (*r_passphrase);
           *r_passphrase = NULL;
         }
+      nvc_release (keymeta);
+      xfree (desc_text_buffer);
       return err;
     }
 
-  buflen = gcry_sexp_canon_len (buf, 0, NULL, NULL);
-  err = gcry_sexp_sscan (&s_skey, &erroff, (char*)buf, buflen);
-  wipememory (buf, buflen);
+  err = sexp_sscan_private_key (result, &erroff, buf);
   xfree (buf);
+  nvc_release (keymeta);
+  xfree (desc_text_buffer);
   if (err)
     {
       log_error ("failed to build S-Exp (off=%u): %s\n",
@@ -1093,213 +1147,8 @@ agent_key_from_file (ctrl_t ctrl, const char *cache_nonce,
           xfree (*r_passphrase);
           *r_passphrase = NULL;
         }
-      return err;
     }
 
-  *result = s_skey;
-  return 0;
-}
-
-
-/* Return the string name from the S-expression S_KEY as well as a
-   string describing the names of the parameters.  ALGONAMESIZE and
-   ELEMSSIZE give the allocated size of the provided buffers.  The
-   buffers may be NULL if not required.  If R_LIST is not NULL the top
-   level list will be stored there; the caller needs to release it in
-   this case.  */
-static gpg_error_t
-key_parms_from_sexp (gcry_sexp_t s_key, gcry_sexp_t *r_list,
-                     char *r_algoname, size_t algonamesize,
-                     char *r_elems, size_t elemssize)
-{
-  gcry_sexp_t list, l2;
-  const char *name, *algoname, *elems;
-  size_t n;
-
-  if (r_list)
-    *r_list = NULL;
-
-  list = gcry_sexp_find_token (s_key, "shadowed-private-key", 0 );
-  if (!list)
-    list = gcry_sexp_find_token (s_key, "protected-private-key", 0 );
-  if (!list)
-    list = gcry_sexp_find_token (s_key, "private-key", 0 );
-  if (!list)
-    {
-      log_error ("invalid private key format\n");
-      return gpg_error (GPG_ERR_BAD_SECKEY);
-    }
-
-  l2 = gcry_sexp_cadr (list);
-  gcry_sexp_release (list);
-  list = l2;
-  name = gcry_sexp_nth_data (list, 0, &n);
-  if (n==3 && !memcmp (name, "rsa", 3))
-    {
-      algoname = "rsa";
-      elems = "ne";
-    }
-  else if (n==3 && !memcmp (name, "dsa", 3))
-    {
-      algoname = "dsa";
-      elems = "pqgy";
-    }
-  else if (n==3 && !memcmp (name, "ecc", 3))
-    {
-      algoname = "ecc";
-      elems = "pabgnq";
-    }
-  else if (n==5 && !memcmp (name, "ecdsa", 5))
-    {
-      algoname = "ecdsa";
-      elems = "pabgnq";
-    }
-  else if (n==4 && !memcmp (name, "ecdh", 4))
-    {
-      algoname = "ecdh";
-      elems = "pabgnq";
-    }
-  else if (n==3 && !memcmp (name, "elg", 3))
-    {
-      algoname = "elg";
-      elems = "pgy";
-    }
-  else
-    {
-      log_error ("unknown private key algorithm\n");
-      gcry_sexp_release (list);
-      return gpg_error (GPG_ERR_BAD_SECKEY);
-    }
-
-  if (r_algoname)
-    {
-      if (strlen (algoname) >= algonamesize)
-        return gpg_error (GPG_ERR_BUFFER_TOO_SHORT);
-      strcpy (r_algoname, algoname);
-    }
-  if (r_elems)
-    {
-      if (strlen (elems) >= elemssize)
-        return gpg_error (GPG_ERR_BUFFER_TOO_SHORT);
-      strcpy (r_elems, elems);
-    }
-
-  if (r_list)
-    *r_list = list;
-  else
-    gcry_sexp_release (list);
-
-  return 0;
-}
-
-
-/* Return true if KEYPARMS holds an EdDSA key.  */
-static int
-is_eddsa (gcry_sexp_t keyparms)
-{
-  int result = 0;
-  gcry_sexp_t list;
-  const char *s;
-  size_t n;
-  int i;
-
-  list = gcry_sexp_find_token (keyparms, "flags", 0);
-  for (i = list ? gcry_sexp_length (list)-1 : 0; i > 0; i--)
-    {
-      s = gcry_sexp_nth_data (list, i, &n);
-      if (!s)
-        continue; /* Not a data element. */
-
-      if (n == 5 && !memcmp (s, "eddsa", 5))
-        {
-          result = 1;
-          break;
-        }
-    }
-  gcry_sexp_release (list);
-  return result;
-}
-
-
-/* Return the public key algorithm number if S_KEY is a DSA style key.
-   If it is not a DSA style key, return 0.  */
-int
-agent_is_dsa_key (gcry_sexp_t s_key)
-{
-  int result;
-  gcry_sexp_t list;
-  char algoname[6];
-
-  if (!s_key)
-    return 0;
-
-  if (key_parms_from_sexp (s_key, &list, algoname, sizeof algoname, NULL, 0))
-    return 0; /* Error - assume it is not an DSA key.  */
-
-  if (!strcmp (algoname, "dsa"))
-    result = GCRY_PK_DSA;
-  else if (!strcmp (algoname, "ecc"))
-    {
-      if (is_eddsa (list))
-        result = 0;
-      else
-        result = GCRY_PK_ECDSA;
-    }
-  else if (!strcmp (algoname, "ecdsa"))
-    result = GCRY_PK_ECDSA;
-  else
-    result = 0;
-
-  gcry_sexp_release (list);
-  return result;
-}
-
-
-/* Return true if S_KEY is an EdDSA key as used with curve Ed25519.  */
-int
-agent_is_eddsa_key (gcry_sexp_t s_key)
-{
-  int result;
-  gcry_sexp_t list;
-  char algoname[6];
-
-  if (!s_key)
-    return 0;
-
-  if (key_parms_from_sexp (s_key, &list, algoname, sizeof algoname, NULL, 0))
-    return 0; /* Error - assume it is not an EdDSA key.  */
-
-  if (!strcmp (algoname, "ecc") && is_eddsa (list))
-    result = 1;
-  else if (!strcmp (algoname, "eddsa")) /* backward compatibility.  */
-    result = 1;
-  else
-    result = 0;
-
-  gcry_sexp_release (list);
-  return result;
-}
-
-
-/* This function returns GPG_ERR_TRUE if S_SKEY represents a shadowed
- * key.  0 is return for other key types.  Any other error may occur
- * if S_SKEY is not valid.  */
-static gpg_error_t
-is_shadowed_key (gcry_sexp_t s_skey)
-{
-  gpg_error_t err;
-  unsigned char *buf;
-  size_t buflen;
-
-  err = make_canon_sexp (s_skey, &buf, &buflen);
-  if (err)
-    return err;
-
-  if (agent_private_key_type (buf) == PRIVATE_KEY_SHADOWED)
-    err = gpg_error (GPG_ERR_TRUE);
-
-  wipememory (buf, buflen);
-  xfree (buf);
   return err;
 }
 
@@ -1319,28 +1168,13 @@ agent_raw_key_from_file (ctrl_t ctrl, const unsigned char *grip,
 
   *result = NULL;
 
-  err = read_key_file (grip, &s_skey, NULL, NULL);
+  err = read_key_file (grip, &s_skey, NULL);
   if (!err)
     *result = s_skey;
   return err;
 }
 
 
-gpg_error_t
-agent_keymeta_from_file (ctrl_t ctrl, const unsigned char *grip,
-                         nvc_t *r_keymeta)
-{
-  gpg_error_t err;
-  gcry_sexp_t s_skey;
-
-  (void)ctrl;
-
-  err = read_key_file (grip, &s_skey, r_keymeta, NULL);
-  gcry_sexp_release (s_skey);
-  return err;
-}
-
-
 /* Return the public key for the keygrip GRIP.  The result is stored
    at RESULT.  This function extracts the public key from the private
    key database.  On failure an error code is returned and NULL stored
@@ -1362,7 +1196,8 @@ agent_public_key_from_file (ctrl_t ctrl,
   const char *uri, *comment;
   size_t uri_length, comment_length;
   int uri_intlen, comment_intlen;
-  char *format, *p;
+  membuf_t format_mb;
+  char *format;
   void *args[2+7+2+2+1]; /* Size is 2 + max. # of elements + 2 for uri + 2
                             for comment + end-of-list.  */
   int argidx;
@@ -1373,7 +1208,7 @@ agent_public_key_from_file (ctrl_t ctrl,
 
   *result = NULL;
 
-  err = read_key_file (grip, &s_skey, NULL, NULL);
+  err = read_key_file (grip, &s_skey, NULL);
   if (err)
     return err;
 
@@ -1404,15 +1239,41 @@ agent_public_key_from_file (ctrl_t ctrl,
   s_skey = NULL;
 
 
-  /* FIXME: The following thing is pretty ugly code; we should
-     investigate how to make it cleaner.  Probably code to handle
-     canonical S-expressions in a memory buffer is better suited for
-     such a task.  After all that is what we do in protect.c.  Need
-     to find common patterns and write a straightformward API to use
-     them.  */
   log_assert (sizeof (size_t) <= sizeof (void*));
 
-  format = xtrymalloc (15+4+7*npkey+10+15+1+1);
+  init_membuf (&format_mb, 256);
+  argidx = 0;
+  put_membuf_printf (&format_mb, "(public-key(%s%%S%%S", algoname);
+  args[argidx++] = &curve;
+  args[argidx++] = &flags;
+  for (idx=0, s=elems; idx < npkey; idx++)
+    {
+      put_membuf_printf (&format_mb, "(%c %%m)", *s++);
+      log_assert (argidx < DIM (args));
+      args[argidx++] = &array[idx];
+    }
+  put_membuf_str (&format_mb, ")");
+  if (uri)
+    {
+      put_membuf_str (&format_mb, "(uri %b)");
+      log_assert (argidx+1 < DIM (args));
+      uri_intlen = (int)uri_length;
+      args[argidx++] = (void *)&uri_intlen;
+      args[argidx++] = (void *)&uri;
+    }
+  if (comment)
+    {
+      put_membuf_str (&format_mb, "(comment %b)");
+      log_assert (argidx+1 < DIM (args));
+      comment_intlen = (int)comment_length;
+      args[argidx++] = (void *)&comment_intlen;
+      args[argidx++] = (void *)&comment;
+    }
+  put_membuf (&format_mb, ")", 2);
+  log_assert (argidx < DIM (args));
+  args[argidx] = NULL;
+
+  format = get_membuf (&format_mb, NULL);
   if (!format)
     {
       err = gpg_error_from_syserror ();
@@ -1425,41 +1286,6 @@ agent_public_key_from_file (ctrl_t ctrl,
       return err;
     }
 
-  argidx = 0;
-  p = stpcpy (stpcpy (format, "(public-key("), algoname);
-  p = stpcpy (p, "%S%S");       /* curve name and flags.  */
-  args[argidx++] = &curve;
-  args[argidx++] = &flags;
-  for (idx=0, s=elems; idx < npkey; idx++)
-    {
-      *p++ = '(';
-      *p++ = *s++;
-      p = stpcpy (p, " %m)");
-      log_assert (argidx < DIM (args));
-      args[argidx++] = &array[idx];
-    }
-  *p++ = ')';
-  if (uri)
-    {
-      p = stpcpy (p, "(uri %b)");
-      log_assert (argidx+1 < DIM (args));
-      uri_intlen = (int)uri_length;
-      args[argidx++] = (void *)&uri_intlen;
-      args[argidx++] = (void *)&uri;
-    }
-  if (comment)
-    {
-      p = stpcpy (p, "(comment %b)");
-      log_assert (argidx+1 < DIM (args));
-      comment_intlen = (int)comment_length;
-      args[argidx++] = (void *)&comment_intlen;
-      args[argidx++] = (void*)&comment;
-    }
-  *p++ = ')';
-  *p = 0;
-  log_assert (argidx < DIM (args));
-  args[argidx] = NULL;
-
   err = gcry_sexp_build_array (&list, NULL, format, args);
   xfree (format);
   for (i=0; array[i]; i++)
@@ -1475,7 +1301,6 @@ agent_public_key_from_file (ctrl_t ctrl,
 }
 
 
-
 /* Check whether the secret key identified by GRIP is available.
    Returns 0 is the key is available.  */
 int
@@ -1503,7 +1328,8 @@ agent_key_available (const unsigned char *grip)
    S-expression.  */
 gpg_error_t
 agent_key_info_from_file (ctrl_t ctrl, const unsigned char *grip,
-                          int *r_keytype, unsigned char **r_shadow_info)
+                          int *r_keytype, unsigned char **r_shadow_info,
+                          unsigned char **r_shadow_info_type)
 {
   gpg_error_t err;
   unsigned char *buf;
@@ -1520,7 +1346,7 @@ agent_key_info_from_file (ctrl_t ctrl, const unsigned char *grip,
   {
     gcry_sexp_t sexp;
 
-    err = read_key_file (grip, &sexp, NULL, NULL);
+    err = read_key_file (grip, &sexp, NULL);
     if (err)
       {
         if (gpg_err_code (err) == GPG_ERR_ENOENT)
@@ -1550,7 +1376,7 @@ agent_key_info_from_file (ctrl_t ctrl, const unsigned char *grip,
           const unsigned char *s;
           size_t n;
 
-          err = agent_get_shadow_info (buf, &s);
+          err = agent_get_shadow_info_type (buf, &s, r_shadow_info_type);
           if (!err)
             {
               n = gcry_sexp_canon_len (s, 0, NULL, NULL);
@@ -1604,7 +1430,7 @@ agent_delete_key (ctrl_t ctrl, const char *desc_text,
   char *default_desc = NULL;
   int key_type;
 
-  err = read_key_file (grip, &s_skey, NULL, NULL);
+  err = read_key_file (grip, &s_skey, NULL);
   if (gpg_err_code (err) == GPG_ERR_ENOENT)
     err = gpg_error (GPG_ERR_NO_SECKEY);
   if (err)
@@ -1701,16 +1527,13 @@ agent_delete_key (ctrl_t ctrl, const char *desc_text,
 
 
 /* Write an S-expression formatted shadow key to our key storage.
- * Shadow key is created by an S-expression public key in PKBUF and
- * card's SERIALNO and the IDSTRING.  With FORCE passed as true an
- * existing key with the given GRIP will get overwritten. If
- * DISPSERIALNO is not NULL the human readable s/n will also be
- * recorded in the key file.   */
+   Shadow key is created by an S-expression public key in PKBUF and
+   card's SERIALNO and the IDSTRING.  With FORCE passed as true an
+   existing key with the given GRIP will get overwritten.  */
 gpg_error_t
 agent_write_shadow_key (const unsigned char *grip,
                         const char *serialno, const char *keyid,
-                        const unsigned char *pkbuf, int force,
-                        const char *dispserialno)
+                        const unsigned char *pkbuf, int force)
 {
   gpg_error_t err;
   unsigned char *shadow_info;
@@ -1737,8 +1560,7 @@ agent_write_shadow_key (const unsigned char *grip,
     }
 
   len = gcry_sexp_canon_len (shdkey, 0, NULL, NULL);
-  err = agent_write_private_key (grip, shdkey, len, force,
-                                 serialno, keyid, dispserialno, 0);
+  err = agent_write_private_key (grip, shdkey, len, force, serialno, keyid, 0);
   xfree (shdkey);
   if (err)
     log_error ("error writing key: %s\n", gpg_strerror (err));
