@@ -50,7 +50,6 @@ get_compliance_cache (enum gnupg_compliance_mode compliance, int for_rng)
   static int r_gnupg   = -1, s_gnupg   = -1;
   static int r_rfc4880 = -1, s_rfc4880 = -1;
   static int r_rfc2440 = -1, s_rfc2440 = -1;
-  static int r_pgp6    = -1, s_pgp6    = -1;
   static int r_pgp7    = -1, s_pgp7    = -1;
   static int r_pgp8    = -1, s_pgp8    = -1;
   static int r_de_vs   = -1, s_de_vs   = -1;
@@ -62,7 +61,6 @@ get_compliance_cache (enum gnupg_compliance_mode compliance, int for_rng)
     case CO_GNUPG:   ptr = for_rng? &r_gnupg   : &s_gnupg  ; break;
     case CO_RFC4880: ptr = for_rng? &r_rfc4880 : &s_rfc4880; break;
     case CO_RFC2440: ptr = for_rng? &r_rfc2440 : &s_rfc2440; break;
-    case CO_PGP6:    ptr = for_rng? &r_pgp6    : &s_pgp6   ; break;
     case CO_PGP7:    ptr = for_rng? &r_pgp7    : &s_pgp7   ; break;
     case CO_PGP8:    ptr = for_rng? &r_pgp8    : &s_pgp8   ; break;
     case CO_DE_VS:   ptr = for_rng? &r_de_vs   : &s_de_vs  ; break;
@@ -541,7 +539,7 @@ gnupg_rng_is_compliant (enum gnupg_compliance_mode compliance)
       if (res == 1)
         {
           char *buf;
-          char *fields[5];
+          const char *fields[5];
 
           buf = gcry_get_config (0, "rng-type");
           if (buf
@@ -590,7 +588,7 @@ gnupg_gcrypt_is_compliant (enum gnupg_compliance_mode compliance)
           /* Libgcrypt might be nice enough to tell us whether it is
            * compliant.  */
           char *buf;
-          char *fields[3];
+          const char *fields[3];
 
           buf = gcry_get_config (0, "compliance");
           if (buf
@@ -623,7 +621,6 @@ gnupg_status_compliance_flag (enum gnupg_compliance_mode compliance)
       return "8";
     case CO_RFC4880:
     case CO_RFC2440:
-    case CO_PGP6:
     case CO_PGP7:
     case CO_PGP8:
       log_assert (!"no status code assigned for this compliance mode");
@@ -675,7 +672,6 @@ gnupg_compliance_option_string (enum gnupg_compliance_mode compliance)
     case CO_GNUPG:   return "--compliance=gnupg";
     case CO_RFC4880: return "--compliance=openpgp";
     case CO_RFC2440: return "--compliance=rfc2440";
-    case CO_PGP6:    return "--compliance=pgp6";
     case CO_PGP7:    return "--compliance=pgp7";
     case CO_PGP8:    return "--compliance=pgp8";
     case CO_DE_VS:   return "--compliance=de-vs";
