@@ -31,9 +31,14 @@
 #include "../common/compliance.h"
 
 
-/* Declaration of a keyserver spec type.  The definition is found in
-   ../common/keyserver.h.  */
-struct keyserver_spec;
+/* Object to hold information pertaining to a keyserver; it also
+   allows building a list of keyservers.  For historic reasons this is
+   not a strlist_t.  */
+struct keyserver_spec
+{
+  struct keyserver_spec *next;
+  char *uri;
+};
 typedef struct keyserver_spec *keyserver_spec_t;
 
 
@@ -259,6 +264,9 @@ struct
     /* The next flag is set internally iff IMPORT_SELF_SIGS_ONLY has
      * been set by the user and is not the default value.  */
     unsigned int expl_import_self_sigs_only:1;
+    /* The next flag is set internally iff IMPORT_CLEAN has
+     * been set by the user and is not the default value.  */
+    unsigned int expl_import_clean:1;
   } flags;
 
   /* Linked list of ways to find a key if the key isn't on the local
