@@ -25,7 +25,6 @@
 #include <errno.h>
 #include <unistd.h>
 #include <time.h>
-#include <assert.h>
 
 #include "gpgsm.h"
 #include <gcrypt.h>
@@ -141,7 +140,7 @@ gpgsm_get_default_cert (ctrl_t ctrl, ksba_cert_t *r_cert)
   int rc;
   char *p;
 
-  hd = keydb_new ();
+  hd = keydb_new (ctrl);
   if (!hd)
     return gpg_error (GPG_ERR_GENERAL);
   rc = keydb_search_first (ctrl, hd);
@@ -161,7 +160,7 @@ gpgsm_get_default_cert (ctrl_t ctrl, ksba_cert_t *r_cert)
           return rc;
         }
 
-      if (!gpgsm_cert_use_sign_p (cert))
+      if (!gpgsm_cert_use_sign_p (cert, 1))
         {
           p = gpgsm_get_keygrip_hexstring (cert);
           if (p)
@@ -218,7 +217,7 @@ get_default_signer (ctrl_t ctrl)
       return NULL;
     }
 
-  kh = keydb_new ();
+  kh = keydb_new (ctrl);
   if (!kh)
     return NULL;
 
@@ -255,7 +254,6 @@ add_certificate_list (ctrl_t ctrl, ksba_cms_t cms, ksba_cert_t cert)
   ksba_cert_ref (cert);
 
   n = ctrl->include_certs;
-  log_debug ("adding certificates at level %d\n", n);
   if (n == -2)
     {
       not_root = 1;
@@ -293,7 +291,7 @@ add_certificate_list (ctrl_t ctrl, ksba_cms_t cms, ksba_cert_t cert)
     }
   ksba_cert_release (cert);
 
-  return rc == -1? 0: rc;
+  return gpg_err_code (rc) == GPG_ERR_NOT_FOUND? 0 : rc;
 
  ksba_failure:
   ksba_cert_release (cert);
@@ -302,6 +300,99 @@ add_certificate_list (ctrl_t ctrl, ksba_cms_t cms, ksba_cert_t cert)
 }
 
 
+#if KSBA_VERSION_NUMBER >= 0x010400 && 0 /* 1.4.0 */
+static gpg_error_t
+add_signed_attribute (ksba_cms_t cms, const char *attrstr)
+{
+  gpg_error_t err;
+  char **fields = NULL;
+  const char *s;
+  int i;
+  unsigned char *der = NULL;
+  size_t derlen;
+
+  fields = strtokenize (attrstr, ":");
+  if (!fields)
+    {
+      err = gpg_error_from_syserror ();
+      log_error ("strtokenize failed: %s\n", gpg_strerror (err));
+      goto leave;
+    }
+
+  for (i=0; fields[i]; i++)
+    ;
+  if (i != 3)
+    {
+      err = gpg_error (GPG_ERR_SYNTAX);
+      log_error ("invalid attribute specification '%s': %s\n",
+                 attrstr, i < 3 ? "not enough fields":"too many fields");
+      goto leave;
+    }
+  if (!ascii_strcasecmp (fields[1], "u"))
+    {
+      err = 0;
+      goto leave; /* Skip unsigned attributes.  */
+    }
+  if (ascii_strcasecmp (fields[1], "s"))
+    {
+      err = gpg_error (GPG_ERR_SYNTAX);
+      log_error ("invalid attribute specification '%s': %s\n",
+                 attrstr, "type is not 's' or 'u'");
+      goto leave;
+    }
+  /* Check that the OID is valid.  */
+  err = ksba_oid_from_str (fields[0], &der, &derlen);
+  if (err)
+    {
+      log_error ("invalid attribute specification '%s': %s\n",
+                 attrstr, gpg_strerror (err));
+      goto leave;
+    }
+  xfree (der);
+  der = NULL;
+
+  if (strchr (fields[2], '/'))
+    {
+      /* FIXME: read from file. */
+    }
+  else /* Directly given in hex.  */
+    {
+      for (i=0, s = fields[2]; hexdigitp (s); s++, i++)
+        ;
+      if (*s || !i || (i&1))
+        {
+          log_error ("invalid attribute specification '%s': %s\n",
+                     attrstr, "invalid hex encoding of the data");
+          err = gpg_error (GPG_ERR_SYNTAX);
+          goto leave;
+        }
+      der = xtrystrdup (fields[2]);
+      if (!der)
+        {
+          err = gpg_error_from_syserror ();
+          log_error ("malloc failed: %s\n", gpg_strerror (err));
+          goto leave;
+        }
+      for (s=fields[2], derlen=0; s[0] && s[1]; s += 2)
+        der[derlen++] = xtoi_2 (s);
+    }
+
+  /* Store the data in the CMS object for all signers.  */
+  err = ksba_cms_add_attribute (cms, -1, fields[0], 0, der, derlen);
+  if (err)
+    {
+      log_error ("invalid attribute specification '%s': %s\n",
+                 attrstr, gpg_strerror (err));
+      goto leave;
+    }
+
+ leave:
+  xfree (der);
+  xfree (fields);
+  return err;
+}
+#endif /*ksba >= 1.4.0 */
+
 
 
 /* Perform a sign operation.
@@ -331,7 +422,7 @@ gpgsm_sign (ctrl_t ctrl, certlist_t signerlist,
 
   audit_set_type (ctrl->audit, AUDIT_TYPE_SIGN);
 
-  kh = keydb_new ();
+  kh = keydb_new (ctrl);
   if (!kh)
     {
       log_error (_("failed to allocate keyDB handle\n"));
@@ -377,10 +468,17 @@ gpgsm_sign (ctrl_t ctrl, certlist_t signerlist,
       goto leave;
     }
 
-  /* We are going to create signed data with data as encap. content */
+  /* We are going to create signed data with data as encap. content.
+   * In authenticode mode we use spcIndirectDataContext instead.  */
   err = ksba_cms_set_content_type (cms, 0, KSBA_CT_SIGNED_DATA);
   if (!err)
-    err = ksba_cms_set_content_type (cms, 1, KSBA_CT_DATA);
+    err = ksba_cms_set_content_type
+      (cms, 1,
+#if KSBA_VERSION_NUMBER >= 0x010400 && 0
+       opt.authenticode? KSBA_CT_SPC_IND_DATA_CTX :
+#endif
+       KSBA_CT_DATA
+       );
   if (err)
     {
       log_debug ("ksba_cms_set_content_type failed: %s\n",
@@ -404,7 +502,7 @@ gpgsm_sign (ctrl_t ctrl, certlist_t signerlist,
 
       /* Although we don't check for ambiguous specification we will
          check that the signer's certificate is usable and valid.  */
-      rc = gpgsm_cert_use_sign_p (cert);
+      rc = gpgsm_cert_use_sign_p (cert, 0);
       if (!rc)
         rc = gpgsm_validate_chain (ctrl, cert, "", NULL, 0, NULL, 0, NULL);
       if (rc)
@@ -430,6 +528,7 @@ gpgsm_sign (ctrl_t ctrl, certlist_t signerlist,
       release_signerlist = 1;
     }
 
+
   /* Figure out the hash algorithm to use. We do not want to use the
      one for the certificate but if possible an OID for the plain
      algorithm.  */
@@ -438,6 +537,11 @@ gpgsm_sign (ctrl_t ctrl, certlist_t signerlist,
   for (i=0, cl=signerlist; cl; cl = cl->next, i++)
     {
       const char *oid;
+      unsigned int nbits;
+      int pk_algo;
+
+      pk_algo = gpgsm_get_key_algo_info (cl->cert, &nbits);
+      cl->pk_algo = pk_algo;
 
       if (opt.forced_digest_algo)
         {
@@ -446,7 +550,21 @@ gpgsm_sign (ctrl_t ctrl, certlist_t signerlist,
         }
       else
         {
-          oid = ksba_cert_get_digest_algo (cl->cert);
+          if (pk_algo == GCRY_PK_ECC)
+            {
+              /* Map the Curve to a corresponding hash algo.  */
+              if (nbits <= 256)
+                oid = "2.16.840.1.101.3.4.2.1"; /* sha256 */
+              else if (nbits <= 384)
+                oid = "2.16.840.1.101.3.4.2.2"; /* sha384 */
+              else
+                oid = "2.16.840.1.101.3.4.2.3"; /* sha512 */
+            }
+          else
+            {
+              /* For RSA we reuse the hash algo used by the certificate.  */
+              oid = ksba_cert_get_digest_algo (cl->cert);
+            }
           cl->hash_algo = oid ? gcry_md_map_name (oid) : 0;
         }
       switch (cl->hash_algo)
@@ -482,27 +600,23 @@ gpgsm_sign (ctrl_t ctrl, certlist_t signerlist,
           goto leave;
         }
 
-      {
-        unsigned int nbits;
-        int pk_algo = gpgsm_get_key_algo_info (cl->cert, &nbits);
+      if (! gnupg_pk_is_allowed (opt.compliance, PK_USE_SIGNING, pk_algo, 0,
+                                 NULL, nbits, NULL))
+        {
+          char  kidstr[10+1];
 
-        if (! gnupg_pk_is_allowed (opt.compliance, PK_USE_SIGNING, pk_algo,
-                                   NULL, nbits, NULL))
-          {
-            char  kidstr[10+1];
+          snprintf (kidstr, sizeof kidstr, "0x%08lX",
+                    gpgsm_get_short_fingerprint (cl->cert, NULL));
+          log_error (_("key %s may not be used for signing in %s mode\n"),
+                     kidstr,
+                     gnupg_compliance_option_string (opt.compliance));
+          err = gpg_error (GPG_ERR_PUBKEY_ALGO);
+          goto leave;
+        }
 
-            snprintf (kidstr, sizeof kidstr, "0x%08lX",
-                      gpgsm_get_short_fingerprint (cl->cert, NULL));
-            log_error (_("key %s may not be used for signing in %s mode\n"),
-                       kidstr,
-                       gnupg_compliance_option_string (opt.compliance));
-            err = gpg_error (GPG_ERR_PUBKEY_ALGO);
-            goto leave;
-          }
-      }
     }
 
-  if (opt.verbose)
+  if (opt.verbose > 1 || opt.debug)
     {
       for (i=0, cl=signerlist; cl; cl = cl->next, i++)
         log_info (_("hash algorithm used for signer %d: %s (%s)\n"),
@@ -513,7 +627,7 @@ gpgsm_sign (ctrl_t ctrl, certlist_t signerlist,
   /* Gather certificates of signers and store them in the CMS object. */
   for (cl=signerlist; cl; cl = cl->next)
     {
-      rc = gpgsm_cert_use_sign_p (cl->cert);
+      rc = gpgsm_cert_use_sign_p (cl->cert, 0);
       if (rc)
         goto leave;
 
@@ -643,6 +757,21 @@ gpgsm_sign (ctrl_t ctrl, certlist_t signerlist,
         }
     }
 
+  /* We can add signed attributes only when build against libksba 1.4.  */
+#if KSBA_VERSION_NUMBER >= 0x010400 && 0 /* 1.4.0 */
+  {
+    strlist_t sl;
+
+    for (sl = opt.attributes; sl; sl = sl->next)
+      if ((err = add_signed_attribute (cms, sl->d)))
+        goto leave;
+  }
+#else
+  if (opt.attributes)
+    log_info ("Note: option --attribute is ignored by this version\n");
+#endif /*ksba >= 1.4.0  */
+
+
   /* We need to write at least a minimal list of our capabilities to
      try to convince some MUAs to use 3DES and not the crippled
      RC2. Our list is:
@@ -678,7 +807,7 @@ gpgsm_sign (ctrl_t ctrl, certlist_t signerlist,
           unsigned char *digest;
           size_t digest_len;
 
-          assert (!detached);
+          log_assert (!detached);
 
           rc = hash_and_copy_data (data_fd, data_md, writer);
           if (rc)
@@ -777,17 +906,23 @@ gpgsm_sign (ctrl_t ctrl, certlist_t signerlist,
                   goto leave;
                 }
               rc = 0;
-              {
-                int pkalgo = gpgsm_get_key_algo_info (cl->cert, NULL);
-                buf = xtryasprintf ("%c %d %d 00 %s %s",
-                                    detached? 'D':'S',
-                                    pkalgo,
-                                    cl->hash_algo,
-                                    signed_at,
-                                    fpr);
-                if (!buf)
-                  rc = gpg_error_from_syserror ();
-              }
+              if (opt.verbose)
+                {
+                  char *pkalgostr = gpgsm_pubkey_algo_string (cl->cert, NULL);
+                  log_info (_("%s/%s signature using %s key %s\n"),
+                            pubkey_algo_to_string (cl->pk_algo),
+                            gcry_md_algo_name (cl->hash_algo),
+                            pkalgostr, fpr);
+                  xfree (pkalgostr);
+                }
+              buf = xtryasprintf ("%c %d %d 00 %s %s",
+                                  detached? 'D':'S',
+                                  cl->pk_algo,
+                                  cl->hash_algo,
+                                  signed_at,
+                                  fpr);
+              if (!buf)
+                rc = gpg_error_from_syserror ();
               xfree (fpr);
               if (rc)
                 {
@@ -813,7 +948,6 @@ gpgsm_sign (ctrl_t ctrl, certlist_t signerlist,
   audit_log (ctrl->audit, AUDIT_SIGNING_DONE);
   log_info ("signature created\n");
 
-
  leave:
   if (rc)
     log_error ("error creating signature: %s <%s>\n",
